#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//GSL
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_math.h>
#define GSL_RANGE_CHECK_OFF


#define _T         1607  //length of the time-series
#define _N         33    //dimension of the series
#define _B         20    //number of bootstraps
#define _nsize     0.05  //nominal size
#define _d         .2    //bandwidth
#define _seed      30
#define _infile    "garchfildata.txt"

gsl_rng *rnd;

// Test Statistic
double Q(gsl_matrix *x1, gsl_matrix *x2, double d)
{
  int t1, t2, n, T=x1->size1, N=x1->size2;
  double Q11, Q12, eqdist, diff, eqdist2;
  
  Q11=Q12=0.0;

  for (t1=0;t1<T;t1++)
  {
     for (t2=0;t2<T;t2++)
     {
       eqdist=eqdist2=0.0;
       for (n=0;n<N;n++)
       {
         diff=gsl_matrix_get(x1,t1,n)-gsl_matrix_get(x1,t2,n);
 	 eqdist+=diff*diff;
         diff=gsl_matrix_get(x1,t1,n)-gsl_matrix_get(x2,t2,n);
	 eqdist2+=diff*diff;
       }
       if (t1!=t2)
	 Q11+=exp(-(double) eqdist/(double) N/2.0/d/d);
       Q12+=exp(-(double) eqdist2/(double) N/2.0/d/d);
       //       printf("%e %e \n",exp(-(double) eqdist/(double) N/2.0/d/d),exp(-(double) eqdist2/(double) N/2.0/d/d));
     }  
  }


  Q11 /= (double) (T-1)*(T-1);
  Q12 /= (double) T*T;
  //  printf("%f %f \n",Q11,Q12);
  return (Q11-2.0*Q12);
}

//empirical CDF transform
void ecdf_transform(gsl_matrix *x)
{
  int i,j;
  gsl_permutation *perm =gsl_permutation_alloc(x->size1);
  gsl_permutation *rank =gsl_permutation_alloc(x->size1);
  gsl_vector_view v;

  for(j=0;j<x->size2;j++)
  {
    v=gsl_matrix_column (x, j); 
    gsl_sort_vector_index (perm,&v.vector);
    gsl_permutation_inverse (rank, perm);
        
    for (i=0;i<x->size1;i++)
      gsl_matrix_set(x,i,j,(double) (gsl_permutation_get(rank,i)+1)/(x->size1+1));
  }
  gsl_permutation_free (perm);
  gsl_permutation_free (rank);
} 

//Generate multivariate gaussian RVs  with correlation ro
void mult_gaussian(gsl_matrix *x, gsl_matrix *chol_ro)
{
  int i,j;

  for (i=0;i<x->size1;i++)
    for (j=0;j<x->size2;j++)
      gsl_matrix_set(x,i,j,gsl_ran_ugaussian(rnd));
  
  //introduce dependence  
  gsl_blas_dtrmm(CblasRight,CblasUpper,CblasNoTrans,CblasNonUnit,1.0,chol_ro,x);

 return;
}

//Generate RV from multivariate gaussian copula
void mult_gaussian_copula(gsl_matrix *x, gsl_matrix *chol_ro)
{
  int i,j;
   
  for (i=0;i<x->size1;i++)
   for (j=0;j<x->size2;j++)
     gsl_matrix_set(x,i,j,gsl_ran_ugaussian(rnd));

  ecdf_transform(x);
  
  for(i=0;i<x->size1;i++)
    for (j=0;j<x->size2;j++)
      gsl_matrix_set(x,i,j,gsl_cdf_ugaussian_Pinv(gsl_matrix_get(x,i,j)));
  
  //introduce dependence  
  gsl_blas_dtrmm(CblasRight,CblasUpper,CblasNoTrans,CblasNonUnit,1.0,chol_ro,x);
  for(i=0;i<x->size1;i++)
    for (j=0;j<x->size2;j++)
      gsl_matrix_set(x,i,j,gsl_cdf_ugaussian_P(gsl_matrix_get(x,i,j)));
}

//Generate mult gaussian RV from copula
void copula_to_mult_gaussian(gsl_matrix *x)
{
  int i,j;

  for(i=0;i<x->size1;i++)
    for (j=0;j<x->size2;j++)
      gsl_matrix_set(x,i,j,gsl_cdf_ugaussian_Pinv(gsl_matrix_get(x,i,j)));
}

//Generate t RVs with correlation ro and nu degrees of freedom
void mult_t(gsl_matrix *x, gsl_matrix *chol_ro, int nu)
{
  int i,j;

  mult_gaussian(x,chol_ro);
  for (i=0;i<x->size1;i++)
    for (j=0;j<x->size2;j++)
      gsl_matrix_set(x,i,j,gsl_matrix_get(x,i,j)*sqrt((double) nu/gsl_ran_chisq(rnd,nu)));
  return;
}

//Generate RV from multivariate t copula
void mult_t_copula(gsl_matrix *x, gsl_matrix *chol_ro,int nu)
{
  int i,j;

  mult_t(x,chol_ro,nu);

   for(i=0;i<x->size1;i++)
    for (j=0;j<x->size2;j++)
      gsl_matrix_set(x,i,j,gsl_cdf_tdist_P(gsl_matrix_get(x,i,j),nu));
}

void standardize(gsl_matrix *x)
{
  int i,j;
  double mean, sd;

  for (j=0;j<x->size2;j++)
  {
    mean=sd=0.0;
    for (i=0;i<x->size1;i++)
    {
      mean+=gsl_matrix_get(x,i,j); 
      sd+=gsl_matrix_get(x,i,j)*gsl_matrix_get(x,i,j);
    }
    mean/=(double) x->size1;
    sd =sqrt(sd/(double) (x->size1-1)-mean*mean*(double) x->size1/(double)(x->size1-1));
    for (i=0;i<x->size1;i++)
      gsl_matrix_set(x,i,j,(gsl_matrix_get(x,i,j)-mean)/sd);
  }
}

//computes the covariance among all elements
void covar(gsl_matrix *x, gsl_matrix *ro)
{
  int i,j;
  gsl_vector_view v1,v2;

  for (i=0;i<x->size2;i++)
    for(j=0;j<x->size2;j++)
    {
      v1=gsl_matrix_column (x, i); 
      v2=gsl_matrix_column (x, j); 
      gsl_matrix_set(ro,i,j,gsl_stats_covariance (v1.vector.data, v1.vector.stride,v2.vector.data,v2.vector.stride,x->size1));
    }
}

//read data
void read_data (gsl_matrix *x)
{
  int i,j;
  float a;
  FILE *fil;  

  fil=fopen(_infile,"rt");

  for (i=0;i<x->size1;i++)
    for(j=0;j<x->size2;j++)
    { 
      fscanf(fil,"%f",&a);
      gsl_matrix_set(x,i,j,a);
    } 
}

int main()
{
  gsl_matrix *x, *x1, *xb, *ro, *ro1;

  double Q0, Q1,p;
  int b,nn,Ip,i,j;

  
  rnd = gsl_rng_alloc(gsl_rng_taus);
  gsl_rng_set (rnd, _seed);

  x=gsl_matrix_alloc(_T,_N);
  x1=gsl_matrix_alloc(_T,_N);
  ro=gsl_matrix_alloc(_N,_N);

  read_data(x);
  ecdf_transform(x);
  gsl_matrix_memcpy(x1,x);
  copula_to_mult_gaussian(x1);

  covar(x1,ro);
  
  gsl_linalg_cholesky_decomp(ro);
  standardize(x);

  mult_gaussian_copula(x1,ro);
  standardize(x1);
 
  Q0=Q(x,x1,_d);
  printf("%f \n",Q0);
 //Parametric bootstrap under the null
  Ip=1;
  for (b=1; b<_B; b++)
  {
    mult_gaussian_copula(x,ro);
    standardize(x);

    Q1=Q(x,x1,_d);
  printf("%f \n",Q1);
  if (Q1>=Q0) // tails are not considered
      Ip++;
  }// end of bootstrap loop
 
  printf("empirical p-value %f \n",(double)Ip/(double)_B);
  return(0);
}
